/* wmlpq - a Windowmaker dockapp to monitor the status of a number 
   of printqueues.
   Copyright (c) 2000,2001 Toby White <tow@theor.ch.cam.ac.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include <dockapp.h>

#include "wmlpq-master.xpm"
Pixmap pixmapshow, pixmapback, mask;

#define WMLPQ_VERSION "0.1"

#define XBORDER 5
#define YBORDER 5
#define LINE_HEIGHT 11
#define CHAR_WIDTH  5
#define CHAR_HEIGHT 7

#define BUFSIZE 512

#define TIME_GRANULARITY 1000
#define LABEL_LENGTH 5
#define NAME_LENGTH 32
#define PATH_LENGTH 256
#define MAX_Q 5

int loopinterval=30;	/* Default rescan interval, in seconds */

char *displayName=NULL, *geometrySpec=NULL, *confFile=NULL;
char *user=NULL;

static DAProgramOption options[] = {
    {"-display", NULL, "display to use", 
     DOString, False, {&displayName} },
    {"-geometry", NULL, "widthxheight+XPOS+YPOS", 
     DOString, False, {&geometrySpec} },
    {"-c", "--config", "path to configuration file", 
     DOString, False, {&confFile} },
};         

enum daemons { lpd=1, lpsched, cups, pdq };

typedef struct queuestuff queue_t;
struct queuestuff 
{
    char label[LABEL_LENGTH+1];	   /* Printed at left; max 5 chars */
    char queuename[NAME_LENGTH+1]; /* Um - name of queue? */
    unsigned char lptype;          /* Type of printer daemon */
    int (*parse_queue)(int);       /* Function pointer to lpq parser */ 
    char queuecmd[PATH_LENGTH+1];  /* Action for checking printqueue */
    int TotalJobs;	               /* Total jobs in queue; up to 8 */
    unsigned int JobList;	       /* bitfield holding 1 for personally 
                                      owned jobs, 0 for other people's */
    time_t last_checked;           /* Time queue was last checked. */
    int loopinterval;	           /* loop interval for this mailbox 
                                      (seconds) */
};

static queue_t queue[MAX_Q];

int init_wmlpq(void);
void do_wmlpq(void);
int parse_lpd(const int q_i);
int parse_lpsched(const int q_i);
int parse_cups(const int q_i);
int parse_pdq(const int q_i);
int parse_fail(const int q_i);
int Read_Config_Int(FILE *fp, const char *conf, int *value);
int Read_Config_String(FILE *fp, const char *conf, char *value);
int Read_Config_File(const char *filename);
void BlitString(const char *name, const int q_i, int new);
void BlitLEDs(const int num_j, char mask, const int q_i, int new);

inline void copyXPMArea(int x, int y, int sx, int sy, int dx, int dy) 
{
    XCopyArea(DADisplay, pixmapback, pixmapshow, DAGC, x, y, sx, sy, dx, dy);
}


int init_wmlpq(void)
{
    int status, q_i;
    unsigned xpmwidth, xpmheight;
    char configFile[PATH_LENGTH+1];

    user=getenv("USER");
    
    for(q_i=0; q_i<MAX_Q; q_i++) 
    {
        queue[q_i].label[0]=0;
        queue[q_i].queuename[0]=0;
        queue[q_i].lptype=0;
        queue[q_i].queuecmd[0]=0;
        queue[q_i].loopinterval=0;
    }

    /* Read config file */
    if (confFile) 
    {
        fprintf(stderr, "Using user-specified config file '%s'.\n", confFile);
        strncpy(configFile, confFile, PATH_LENGTH);
    }
    else if (snprintf(configFile, PATH_LENGTH, "%s/.wmlpqrc", getenv("HOME"))<0)
    {
        fprintf(stderr,"Path of cwd too long.");
        exit(0);
    }

    if (!Read_Config_File(configFile))
    {
        printf("Error parsing %s\n",configFile);
    }
    
/* Put in any defaults necessary. */
    for (q_i=0;q_i<MAX_Q;q_i++)
    {
        if (!queue[q_i].label[0]) queue[q_i].lptype=0;
        if (!queue[q_i].loopinterval) queue[q_i].loopinterval=loopinterval;
        if (!queue[q_i].queuecmd[0])
        {
            switch (queue[q_i].lptype)
            {
                case lpd:
                    strncpy(queue[q_i].queuecmd,"lpq\000",PATH_LENGTH);
                    break;
                case lpsched:
                    strncpy(queue[q_i].queuecmd,"lpstat\000",PATH_LENGTH);
                    break;
                case cups:
                    strncpy(queue[q_i].queuecmd,"lpstat\000",PATH_LENGTH);
                    break;
                case pdq:
                    strncpy(queue[q_i].queuecmd,"lpd_status\000",PATH_LENGTH);
                    break;
                default:
                    strncpy(queue[q_i].queuecmd,"false\000",PATH_LENGTH);
                    break;
            }
        }
        
        switch (queue[q_i].lptype)
        {
            case lpd:
                queue[q_i].parse_queue=&parse_lpd;
                break;
            case lpsched:
                queue[q_i].parse_queue=&parse_lpsched;
                break;
            case cups:
                queue[q_i].parse_queue=&parse_cups;
                break;
            case pdq:
                queue[q_i].parse_queue=&parse_pdq;
                break;
            default:
                queue[q_i].parse_queue=&parse_fail;
                break;
        }
    }

    if (queue[0].lptype==0)
    {
        char *m=NULL;
        if ( (m=getenv("LPDEST")) ) 
        {
            strncpy(queue[0].queuename, m, LABEL_LENGTH);
            strncpy(queue[0].label, m, LABEL_LENGTH);
            queue[0].lptype=lpsched;
            strncpy(queue[0].queuecmd,"lpstat\000",PATH_LENGTH);
            queue[0].parse_queue=&parse_lpsched;
        }
        else if ( (m=getenv("PRINTER")) ) 
        {
            strncpy(queue[0].queuename, m, LABEL_LENGTH);
            strncpy(queue[0].label, m, LABEL_LENGTH);
            queue[0].lptype=lpd;
            strncpy(queue[0].queuecmd,"lpq\000",PATH_LENGTH);
            queue[0].parse_queue=&parse_lpd;
        }
        else  
        {
            strncpy(queue[0].queuename,"lp\00",NAME_LENGTH);
            strncpy(queue[0].label,"lp\000",LABEL_LENGTH);
            queue[0].lptype=lpd;
            strncpy(queue[0].queuecmd,"lpq\000",PATH_LENGTH);
            queue[0].parse_queue=&parse_lpd;
            fprintf(stderr, 
                    "Cannot use '%s' nor the PRINTER environment variable.\n", 
                    configFile);
        }
    }

    status=DAMakePixmapFromData(wmlpq_master_xpm, &pixmapback, &mask, 
                                &xpmwidth, &xpmheight);
    if (!status) abort();
    pixmapshow=XCreatePixmap(DADisplay, DAWindow, 54, 54, DADepth);
    XCopyArea(DADisplay, pixmapback, pixmapshow, 
              DAGC, XBORDER, YBORDER, 54, 54, 0, 0);
    
    for (q_i=0;q_i<MAX_Q;q_i++)
    {
        if (queue[q_i].label[0])
            BlitString(queue[q_i].label, q_i, 0);
    }

    DASetPixmap(pixmapshow);
    
    for (q_i=0;q_i<MAX_Q;q_i++)
    {
        status=queue[q_i].parse_queue(q_i);
        if (status) 
        {
            /* handle error */
            /* continue; */
        }
        BlitLEDs(queue[q_i].TotalJobs, queue[q_i].JobList, q_i, 0);
    }

/*     DASetCallbacks(callbacks); */
/* Implement as and when everything else works. */    
    
    return 1;
}

void do_wmlpq(void)
{
    for (;;)
    {
        XEvent event;
        size_t q_i;
        if ( DANextEventOrTimeout(&event, TIME_GRANULARITY) ) 
            DAProcessEvent(&event);

        for (q_i=0;q_i<MAX_Q;q_i++)
        {
            int status;
            
            time_t now=time(NULL);
            if ( difftime(now, queue[q_i].last_checked) 
                 > (double)queue[q_i].loopinterval )
            {
                status=queue[q_i].parse_queue(q_i);
                queue[q_i].last_checked=now;
                if (status) 
                {
                    /* handle error */
                    /* continue; */
                }
                BlitLEDs(queue[q_i].TotalJobs, queue[q_i].JobList, q_i, 0);
            }
        }
        DASetPixmap(pixmapshow);
    }    
}

int parse_lpd(const int q_i)
{
    FILE *lpq_out;
    char buffer[80+1];
    char owner[NAME_LENGTH+1];
    char lpqcmd[PATH_LENGTH+3+NAME_LENGTH+1];
    unsigned int rList;
    int iter,total;

    strncpy(lpqcmd,queue[q_i].queuecmd,PATH_LENGTH+1);
    strncat(lpqcmd," -P",3);
    strncat(lpqcmd,queue[q_i].queuename,NAME_LENGTH);

    lpq_out=popen(lpqcmd,"r");
            
    total=queue[q_i].TotalJobs=0;
    rList=queue[q_i].JobList=0;
    
    if (fgetc(lpq_out)==EOF) return pclose(lpq_out);
    
    fscanf(lpq_out, "%80[^\n]%*[^\n]", buffer);
    if (strncmp("ank",buffer,3)) return pclose(lpq_out);
    /* Discard first line */

    for (iter=0;iter<8;iter++)
    {
        rList<<=1;
        if (fscanf(lpq_out, "%*s %s %*[^\n]", owner)==EOF ) continue;
        total++;
        /* Check that it is actually the right format? */
        if (!strcmp(user, owner)) rList|=01;
    }
    
    
    for (iter=8;iter>0;iter--)
    {
        queue[q_i].JobList>>=1;
        if (rList&0200) queue[q_i].JobList|=0200;
        rList<<=1;
    }

    queue[q_i].TotalJobs=total;
    
    return pclose(lpq_out);
}

int parse_lpsched(const int q_i)
{
    FILE *lpq_out;
    char buffer[80+1];
    char *qname=queue[q_i].queuename;
    char owner[NAME_LENGTH+1];
    char lpqcmd[PATH_LENGTH+6+NAME_LENGTH+1];
    unsigned int rList;
    int iter,total;

    strncpy(lpqcmd, queue[q_i].queuecmd, PATH_LENGTH+1 );
    strncat(lpqcmd," -u -p",6);
    strncat(lpqcmd, qname, NAME_LENGTH );

    lpq_out=popen(lpqcmd,"r");
            
    total=0;
    rList=0;
    
    for (iter=0;iter<8;iter++)
    {
        rList<<=1;
        if ( fscanf(lpq_out, "%s %s %*[^\n]", buffer, owner)==EOF ) continue;
        if ( !strncmp("printer", buffer, 7) ) continue;
        total++;
        if ( !strcmp(user, owner) ) rList|=01;
    }
    
    queue[q_i].JobList=0;
    
    for (iter=8;iter>0;iter--)
    {
        queue[q_i].JobList>>=1;
        if ( rList & 0200 ) queue[q_i].JobList|=0200;
        rList<<=1;
    }

    queue[q_i].TotalJobs=total;
    
    return pclose(lpq_out);
}

int parse_cups(const int q_i)
{
    /*Untested*/
    FILE *lpq_out;
    char buffer[80+1];
    char *qname=queue[q_i].queuename;
    char owner[NAME_LENGTH+1];
    char lpqcmd[PATH_LENGTH+6+NAME_LENGTH+1];
    unsigned int rList;
    int iter,total;

    strncpy(lpqcmd, queue[q_i].queuecmd, PATH_LENGTH+1 );
    strncat(lpqcmd," -p",6);
    strncat(lpqcmd, qname, NAME_LENGTH );

    lpq_out=popen(lpqcmd,"r");
            
    total=0;
    rList=0;
    
    for (iter=0;iter<8;iter++)
    {
        rList<<=1;
        if ( fscanf(lpq_out, "%s %s %*[^\n]", buffer, owner)==EOF ) continue;
        if ( !strncmp(qname, buffer, 7) ) continue;
        total++;
        if ( !strcmp(user, owner) ) rList|=01;
    }
    
    queue[q_i].JobList=0;
    
    for (iter=8;iter>0;iter--)
    {
        queue[q_i].JobList>>=1;
        if ( rList & 0200 ) queue[q_i].JobList|=0200;
        rList<<=1;
    }

    queue[q_i].TotalJobs=total;
    
    return pclose(lpq_out);
}

int parse_pdq(const int q_i)
{
    /*Untested*/
    FILE *lpq_out;
    char buffer[80+1];
    char owner[NAME_LENGTH+1];
    char lpqcmd[PATH_LENGTH+3+NAME_LENGTH+1];
    unsigned int rList;
    int iter,total;

    strncpy(lpqcmd,queue[q_i].queuecmd,PATH_LENGTH+1);
    strcat(lpqcmd," \000");
    strncat(lpqcmd,queue[q_i].queuename,NAME_LENGTH);

    lpq_out=popen(lpqcmd,"r");
            
    total=queue[q_i].TotalJobs=0;
    rList=queue[q_i].JobList=0;
    
    if (fgetc(lpq_out)==EOF) return pclose(lpq_out);
    
    fscanf(lpq_out, "%80[^\n]%*[^\n]", buffer);
    if (strncmp("ank",buffer,3)) return pclose(lpq_out);
    /* Discard first line */

    for (iter=0;iter<8;iter++)
    {
        rList<<=1;
        if (fscanf(lpq_out, "%*s %s %*[^\n]", owner)==EOF ) continue;
        total++;
        /* Check that it is actually the right format? */
        if (!strcmp(user, owner)) rList|=01;
    }
    
    
    for (iter=8;iter>0;iter--)
    {
        queue[q_i].JobList>>=1;
        if (rList&0200) queue[q_i].JobList|=0200;
        rList<<=1;
    }

    queue[q_i].TotalJobs=total;
    
    return pclose(lpq_out);
}

int parse_fail(const int q_i)
{
    /*For crapness' sake. */
    queue[q_i].TotalJobs=0;
    queue[q_i].JobList=0;
    return 0;
}


/* Blits a string at given co-ordinates
   If a ``new'' parameter is given, all digits will be yellow
*/
void BlitString(const char *name, const int q_i, int new)
{
    size_t l_i;
    int	c, k = 0, y=LINE_HEIGHT*q_i;

    for(l_i=0; l_i<LABEL_LENGTH; l_i++) 
    {
        c = toupper(name[l_i]); 
        if(c >= 'A' && c <= 'Z') 
        { /* Cannot use isalpha() because we may be outside a C-like locale.*/
            c -= 'A';
            copyXPMArea(c * (CHAR_WIDTH+1), 74, CHAR_WIDTH+1, CHAR_HEIGHT+1, k, y); 
            /* Move characters around in master.xpm */
            k += CHAR_WIDTH+1;
        }
        else 
        {   /* it's a number or symbol */
            c -= '0';
            if ( new )
                copyXPMArea(c * (CHAR_WIDTH+1) + 65, 0, CHAR_WIDTH+1, 
                            CHAR_HEIGHT+1, k, y);
            else
                copyXPMArea(c * (CHAR_WIDTH+1), 64, CHAR_WIDTH+1, 
                            CHAR_HEIGHT+1, k, y);
            k += (CHAR_WIDTH+1);
        }
    }
}

void BlitLEDs(const int num_j, char mask, const int q_i, int new)
{
    int colour, red=81, green=87, blank=93;
    int count=0, height=5, xstep=29;

    while ( count++ < 8 ) 
    {
        if (count>num_j)
            colour=blank;
        else if (mask & 01) /* This job is ours */
        {
            colour=green;
        }
        else colour=red;
        mask >>= 1;
        xstep=height?xstep+3:xstep+2;
        height=height?0:5;
        copyXPMArea(colour, 66, 4, 4, xstep, 1 + height + q_i * LINE_HEIGHT);
    }
}

int Read_Config_String(FILE *fp, const char *conf, char *value)
{
    char buf[BUFSIZE];
    char *p=NULL;

    if (!fp) return 0;

    rewind(fp);
    
    while(!feof(fp))
    {
        int len;
        if( !fgets(buf, BUFSIZE-1, fp) )
            break;
        len=strlen(buf);
        if( buf[len-1] == '\n' ) buf[len-1] = 0;        /* Strip linefeed */
        for( p=buf; *p && *p != '#'; p++ ); 
        *p = 0;                                         /* Strip comments */
        for( p=buf; *p && isspace(*p); p++);            /* Skip leading spaces */

        if( !strncmp(p, conf, strlen(conf)) )
        {
            for( p+=strlen(conf); *p && isspace(*p); p++);
            strcpy(value,p);
            return 1;
        }
    }
    return 0;
}

int Read_Config_Int(FILE *fp, const char *conf, int *value)
{
    char buf[BUFSIZE];
    
    if (Read_Config_String(fp, conf, buf))
    {
        *value=(int)strtol(buf,NULL,0);
        return 1;
    }
    
    return 0;
}

int Read_Config_File(const char *filename)
{
    FILE *fp;
    char lptype[8], qcmd[PATH_LENGTH];
    int q_i=0;
    
    fp = fopen(filename, "r");
    if (!fp) 
    {
        fprintf(stderr,"Unable to open '%s'.\n",filename);
        return 0;
    }

    if (Read_Config_String(fp, "lptype ", lptype))
    {
        int lpt=lpd;
        if (!strcmp(lptype,"lpd"))
            lpt=lpd;
        else if (!strcmp(lptype,"lpsched"))
            lpt=lpsched;
        else if (!strcmp(lptype,"cups"))
            lpt=cups;
        else    
        {
            fprintf(stderr,"%s not supported\n",lptype);
        }
        for (q_i=0;q_i<MAX_Q;q_i++)
            queue[q_i].lptype=lpt;
    }
        
    if (Read_Config_String(fp, "queuecmd ", qcmd))
    {
        for (q_i=0;q_i<MAX_Q;q_i++)
            strncpy(queue[q_i].queuecmd,qcmd,128);
    }
    
    for (q_i=0;q_i<MAX_Q;q_i++)
    {
        char digit[2], optionname[15];
        snprintf(digit,2,"%d",q_i);
        
        strncpy(optionname,"label.",13);
        Read_Config_String(fp, strcat(optionname,digit), queue[q_i].label);
        strncpy(optionname,"queuename.",13);
        Read_Config_String(fp, strcat(optionname,digit), queue[q_i].queuename);
        strncpy(optionname,"interval.",13);
        Read_Config_Int(fp, strcat(optionname,digit), &(queue[q_i].loopinterval));
    }
    
    return 1;
}


int main(int argc, char *argv[])
{
    DAParseArguments(argc, argv, options, 3,\
                     "Printqueue monitor dockapp", "wmlpq "WMLPQ_VERSION);

    DAInitialize(displayName, "wmlpq", 54, 54, argc, argv);
    
    init_wmlpq();
    
    DAShow();
    
    do_wmlpq();
    
    return 1;
}
